#!/usr/bin/env python


import logging
import random
import threading
import time
import unittest
from test.test_helper import Probe, Poller, ExecutionProbe, TestHelper
from plico.utils.executor import Task, SingleThreadExecutor,\
    SingleThreadImmediateExecutor, BusyExecutorException, Executor,\
    DoneTaskControl, FunctionTask, MultiThreadExecutor, FunctionTaskException,\
    TaskTimeoutError, logTaskFailure, TaskDelayingExecutor
from plico.utils.decorator import override, synchronized
from plico.utils.logger import DummyLogger, LoggerListener, ObservableLogger



__version__= "$Id: executor_test.py 26 2018-01-26 19:06:25Z lbusoni $"


class TaskCompletionProbe(Probe):

    def __init__(self, task):
        super(TaskCompletionProbe, self).__init__()
        self._task= task

    def probe(self):
        pass

    def isSatisfied(self):
        return self._task.hasBeenPerformed()

    def errorMessage(self):
        return "Task has not been performed"


class RunningNumberGenerator(object):

    def __init__(self):
        self._nextSeqNo= 0
        self._lock= threading.RLock()

    def nextNumber(self):
        self._lock.acquire()
        result= self._nextSeqNo
        self._nextSeqNo+= 1
        self._lock.release()

        return result


class DummyTask(Task):

    def __init__(self, sequenceNumberGenerator=RunningNumberGenerator(),
                 name="dummy"):
        super(DummyTask, self).__init__()
        self._sequenceNumberGenerator= sequenceNumberGenerator
        self._name= name

        self._runningNumber= None
        self._performed= False
        self._mutex= threading.Lock()


    @override
    def name(self):
        return self._name


    @synchronized("_mutex")
    @override
    def perform(self):
        assert not self._performed
        self._runningNumber= self._sequenceNumberGenerator.nextNumber()
        self._performed= True


    @synchronized("_mutex")
    def hasBeenPerformed(self):
        return self._performed


    def runningNumber(self):
        assert self._runningNumber is not None
        return self._runningNumber


class FailingTask(Task):

    def __init__(self, name=None):
        super(FailingTask, self).__init__()
        self._name= name
        self._taskPerformed= False
        self._mutex= threading.Lock()

    @synchronized("_mutex")
    @override
    def perform(self):
        self._taskPerformed= True
        raise Exception("requested failure")


    @synchronized("_mutex")
    def hasBeenPerformed(self):
        return self._taskPerformed


    @override
    def name(self):
        if self._name is None:
            return "FailingTask"
        else:
            return self._name


class SlowTask(Task):

    def __init__(self, sleepDurationSec, name=""):
        Task.__init__(self)
        self._sleepDurationSec= sleepDurationSec
        self._name= name
        self._performed= False
        self._busy= False
        self._canceled= False
        self._mutex= threading.RLock()


    @override
    def perform(self):
        assert not self._performed

        self._setBusyFlag(True)
        MAX_SLEEP_TIME_PER_PASS_SEC= 0.5
        remainingSleepTimeSec= self._sleepDurationSec
        while remainingSleepTimeSec > 0.0 and not self.isCanceled():
            sleepTimeSec= min(MAX_SLEEP_TIME_PER_PASS_SEC,
                              self._sleepDurationSec)
            time.sleep(sleepTimeSec)
            remainingSleepTimeSec-= sleepTimeSec
        self._setBusyFlag(False)
        self._onPerformed()


    @override
    def name(self):
        return self._name


    @synchronized("_mutex")
    def _onPerformed(self):
        self._performed= True


    @synchronized("_mutex")
    def _setBusyFlag(self, busy):
        self._busy= busy


    @synchronized("_mutex")
    def isBusy(self):
        return self._busy


    @synchronized("_mutex")
    def hasBeenPerformed(self):
        return self._performed


    @synchronized("_mutex")
    def isCanceled(self):
        return self._canceled


    @synchronized("_mutex")
    def cancel(self):
        self._canceled= True


class SingleThreadExecutorTest(unittest.TestCase):


    def setUp(self):
        logger= DummyLogger()
        self.executor= SingleThreadExecutor(logger)
        self._seqNoGenerator= RunningNumberGenerator()


    def test_optimistic(self):
        t= self._createDummyTask()
        self.executor.execute(t)
        Poller(5).check(TaskCompletionProbe(t))
        self.assertTrue(t.hasBeenPerformed())


    def test_pessimistic_task(self):
        task_1= FailingTask()
        task_2= self._createDummyTask()
        self.executor.execute(task_1)
        self.executor.execute(task_2)
        Poller(5).check(TaskCompletionProbe(task_1))
        Poller(5).check(TaskCompletionProbe(task_2))


    def test_FIFO_like_behaviour(self):
        task_1= self._createDummyTask()
        task_2= self._createDummyTask()
        task_3= self._createDummyTask()
        self.executor.execute(task_1)
        self.executor.execute(task_2)
        self.executor.execute(task_3)
        Poller(5).check(TaskCompletionProbe(task_1))
        Poller(5).check(TaskCompletionProbe(task_2))
        Poller(5).check(TaskCompletionProbe(task_3))
        self.assertTrue(task_1.runningNumber() < task_2.runningNumber())
        self.assertTrue(task_2.runningNumber() < task_3.runningNumber())


    def _createDummyTask(self):
        return DummyTask(self._seqNoGenerator)


    def test_should_return_task_control(self):
        class LongTask(Task):

            def __init__(self, sleepDurationSec):
                self._sleepDurationSec= sleepDurationSec
                self._done= False

            @override
            def perform(self):
                while not self._done:
                    time.sleep(self._sleepDurationSec)

            def terminate(self):
                self._done= True

        task= LongTask(0.01)
        taskCtrl= self.executor.execute(task)
        self.assertFalse(taskCtrl.isDone())
        Poller(1, pollingDelaySec=0.01).check(ExecutionProbe(
            lambda: self.assertTrue(taskCtrl.isRunning())))
        task.terminate()
        Poller(2, pollingDelaySec=0.01).check(
            ExecutionProbe(lambda: self.assertTrue(taskCtrl.isDone())))

        task= LongTask(0.5)
        taskCtrl= self.executor.execute(task)
        task.terminate()
        taskCtrl.waitForCompletion()
        self.assertTrue(taskCtrl.isDone())


class SingleThreadImmediateExecutorTest(unittest.TestCase):


    def setUp(self):
        logger= DummyLogger()
        self.wrappedExecutor= SingleThreadExecutor(logger)
        self.executor= SingleThreadImmediateExecutor(self.wrappedExecutor)


    def test_construction(self):
        self.assertTrue(self.executor.isIdle())


    def test_should_discard_task_when_executor_is_busy(self):

        class SlowTaskBusyProbe(Probe):
            def __init__(self, task_1):
                Probe.__init__(self)
                self._task= task_1

            @override
            def probe(self):
                pass

            @override
            def isSatisfied(self):
                return self._task.isBusy()

            @override
            def errorMessage(self):
                return "Task is not busy"


        class ExecutorIdleProbe(Probe):
            def __init__(self, executor):
                Probe.__init__(self)
                self.executor= executor

            @override
            def probe(self):
                pass

            @override
            def isSatisfied(self):
                return self.executor.isIdle()

            @override
            def errorMessage(self):
                return "Executor is busy"


        SLEEP_DURATION_SEC= 1.0
        task_1= SlowTask(sleepDurationSec=SLEEP_DURATION_SEC, name="one")

        self.executor.execute(task_1)
        Poller(4).check(SlowTaskBusyProbe(task_1))
        time.sleep(SLEEP_DURATION_SEC / 5.0)

        task_2= DummyTask("two")
        try:
            self.executor.execute(task_2)
        except BusyExecutorException as e:
            self.assertEqual("one", e.nameOfCurrentTask)

        Poller(4).check(ExecutorIdleProbe(self.executor))
        self.executor.execute(task_2)


    def test_should_ignore_failing_tasks(self):
        failingTask= FailingTask("broken")
        self.executor.execute(failingTask)
        Poller(4).check(TaskCompletionProbe(failingTask))


        dummyTask= SlowTask(0, "slow task")
        Poller(1).check(ExecutionProbe(lambda:
                        self.executor.execute(dummyTask)))
        Poller(4).check(TaskCompletionProbe(dummyTask))


    def test_thread_safetyness(self):
        if TestHelper.areLongRunningTestsInhibited():
            logging.warning("Skipping test because"
                            " long running tests are inhibited")
            return

        class UnsynchronziedSerialExecutor(Executor):
            def __init__(self):
                Executor.__init__(self)
                self._nInvocations= 0


            @override
            def execute(self, task):
                newValue= self._nInvocations + 1
                task.perform()
                self._nInvocations= newValue
                return DoneTaskControl()


            def nInvocations(self):
                return self._nInvocations


        class StressThread(threading.Thread):

            def __init__(self, executor, task, nPasses):
                threading.Thread.__init__(self)
                self._executor= executor
                self._task= task
                self._nPasses= nPasses
                self._nSuccessfullInvocations= 0


            def run(self):
                logging.debug("%s: run" % t.name)
                for _ in range(0, self._nPasses):
                    self._nSuccessfullInvocations+= 1
                    try:
                        self._executor.execute(self._task)
                    except:
                        time.sleep(random.random()* 0.01)
                        self._nSuccessfullInvocations-= 1


            def nSuccessFullInvocations(self):
                return self._nSuccessfullInvocations

        self.wrappedExecutor= UnsynchronziedSerialExecutor()
        self.executor= SingleThreadImmediateExecutor(self.wrappedExecutor)

        N_THREADS= 8
        N_PASSES= 1000
        threads= []

        for _ in range(0, N_THREADS):
            task= SlowTask(0)
            t= StressThread(self.executor, task, N_PASSES)
            threads.append(t)

        for t in threads:
            logging.debug("Starting " + t.name)
            t.start()

        totalSuccessFullInvocations= 0
        for t in threads:
            logging.debug("Waiting until all threads are done")
            t.join()
            logging.debug("#successFullinvocations(%s): %d" %(
                t.name, t.nSuccessFullInvocations()))
            totalSuccessFullInvocations+= t.nSuccessFullInvocations()

        self.assertEqual(totalSuccessFullInvocations,
                         self.wrappedExecutor.nInvocations())


class FunctionTaskTest(unittest.TestCase):


    def _nulladicFunc(self):
        self._funcInvoked= True


    def test_should_execute_nulladic_function(self):
        self._funcInvoked= False
        t= FunctionTask(self._nulladicFunc)
        t.perform()
        self.assertTrue(self._funcInvoked)


    def test_should_have_name(self):
        t= FunctionTask(self._nulladicFunc)
        self.assertTrue("_nulladicFunc" in t.name())


    def test_should_have_custom_name(self):
        t= FunctionTask(self._nulladicFunc).withName("spam")
        self.assertEqual("spam", t.name())


    def test_should_execute_bounded_method(self):
        if TestHelper.areLongRunningTestsInhibited():
            return

        class Tux(object):

            def fly(self, *args, **kwds):
                self._args= args
                self._kwds= kwds
                self._funcInvoked= True


        executor= MultiThreadExecutor(
            maxThreads=10, logger=DummyLogger())
        tuxes= []
        futures= []
        for _ in range(0, 1000):
            tux= Tux()
            tuxes.append(tux)
            t= FunctionTask(tux.fly, 1, 2, "foobar", foo="bar")
            future= executor.execute(t)
            futures.append(future)

        for each in futures:
            each.waitForCompletion()

        for eachTux in tuxes:
            self.assertEqual((1, 2, "foobar"), eachTux._args)
            self.assertEqual({"foo": "bar"}, eachTux._kwds)
            self.assertTrue(eachTux._funcInvoked)


    def _method(self, l, d, lock):
        with lock:
            d["foo"]= 1
            l.append("foo")


    def test_should_execute_self_method(self):
        if TestHelper.areLongRunningTestsInhibited():
            return

        executor= MultiThreadExecutor(
            maxThreads=10, logger=DummyLogger())
        N_TASKS= 5000
        results= []
        d= {}
        lock= threading.Lock()
        futures= []
        for _ in range(0, N_TASKS):
            t= FunctionTask(self._method, results, d, lock)
            future= executor.execute(t)
            futures.append(future)

        for each in futures:
            each.waitForCompletion()

        self.assertEqual(N_TASKS, len(results))
        for each in results:
            self.assertEqual("foo", each)


    def test_should_provide_exception(self):
        class CustomException(Exception):
            pass

        def fail():
            raise CustomException("don't panic")
        task= FunctionTask(fail)
        self.assertRaises(CustomException, task.perform)
        self.assertEqual("don't panic", str(task.getException()))
        self.assertRaises(CustomException, task.reraiseExceptionIfAny)

        def okay():
            pass
        task= FunctionTask(okay)
        task.perform()
        self.assertTrue(task.getException() is None)
        task.reraiseExceptionIfAny()


    def test_should_provide_result(self):
        def f():
            return 42
        task= FunctionTask(f)
        self.assertRaises(FunctionTaskException, task.getResult)
        task.perform()
        self.assertEqual(42, task.getResult())


class MultiThreadExecutorTest(unittest.TestCase):


    def setUp(self):
        logger= DummyLogger()
        self.executor= MultiThreadExecutor(3, logger)


    def test_should_use_multiple_threads(self):
        if TestHelper.areLongRunningTestsInhibited():
            return

        sleepDurationSec= 2
        f1= self.executor.execute(SlowTask(sleepDurationSec))
        f2= self.executor.execute(SlowTask(sleepDurationSec))
        f3= self.executor.execute(SlowTask(sleepDurationSec))

        upperExecutionLimit= 1.5 * sleepDurationSec
        t0= time.time()
        for each in [f1, f2, f3]:
            each.waitForCompletion()
        t1= time.time()
        self.assertTrue(t1 - t0 < upperExecutionLimit)



    def test_should_raise_timeout_error(self):
        task= SlowTask(sleepDurationSec=100.0)
        taskCtrl= self.executor.execute(task)
        self.assertRaises(TaskTimeoutError,
                          taskCtrl.waitForCompletion, timeoutSec=0.01)
        task.cancel()


class LogTaskFailureDecoratorTest(unittest.TestCase):

    def test_logging(self):

        class MyLoggerListener(LoggerListener):

            def __init__(self):
                self.errorCount= 0


            def onError(self, message):
                self.errorCount+= 1


        class TestTask(Task):
            def __init__(self, logger):
                Task.__init__(self)
                self._logger= logger

            @logTaskFailure
            def perform(self):
                raise Exception("expected failure")


        logger= ObservableLogger(DummyLogger())
        listener= MyLoggerListener()
        self.assertRaises(Exception,
                          TestTask(logger).perform)
        self.assertTrue(1, listener.errorCount)


class TaskDelayingExecutorTest(unittest.TestCase):

    def test_delay(self):
        class FooTask(Task):
            def __init__(self, results):
                super(FooTask, self).__init__()
                self._results= results

            @override
            def perform(self):
                self._results.append(23)

        taskResult= []
        task= FooTask(taskResult)
        exe= TaskDelayingExecutor()
        taskControl= exe.execute(task)

        tasks= exe.getDelayedTasks()
        self.assertEqual(1, len(tasks))
        self.assertEqual(0, len(taskResult))
        self.assertFalse(taskControl.isDone())
        self.assertFalse(taskControl.isRunning())

        exe.executeDelayedTasks()
        self.assertEqual(23, taskResult[0])
        self.assertTrue(taskControl.isDone())


if __name__ == '__main__':
    logging.basicConfig(level=logging.DEBUG)
    unittest.main()



from plico.rpc.abstract_remote_procedure_call import \
    AbstractRemoteProcedureCall
from plico.utils.timeout import Timeout


__version__= "$Id: serverinfo_client.py 35 2018-01-28 13:33:16Z lbusoni $"


class ServerInfoClient(object):


    def __init__(self,
                 rpcHandler,
                 requestSocket,
                 logger):
        assert isinstance(rpcHandler, AbstractRemoteProcedureCall)

        self._rpcHandler= rpcHandler
        self._requestSocket= requestSocket
        self._logger= logger



    def serverInfo(self, timeoutInSec=Timeout.SERVER_INFO):
        return self._rpcHandler.sendRequest(
            self._requestSocket, 'serverInfo',
            [], timeout=timeoutInSec)



#!/usr/bin/env python
import unittest
import time
import io
import os
import logging
import subprocess
from test.test_helper import MessageInFileProbe, Poller, TestHelper
import shutil
import shlex
from plico.utils.logger import Logger
from plico.rpc.zmq_remote_procedure_call import ZmqRemoteProcedureCall,\
    ZmqRpcTimeoutError



class MyException(Exception):
    pass


class HasALogger():
    def __init__(self, value):
        self.logger= Logger.of('HasALogger')
        self.value= value


    def __getstate__(self):
        self.logger= None
        return self.__dict__


    def __setstate__(self, d):
        self.__dict__ = d
        self.logger= Logger.of('HasANewLogger')


class MyObject():

    def __init__(self, value):
        self._value= value


    def getValue(self):
        return self._value


class MyPublisher():
    REPLY_PORT= 4011
    PUBLISHER_PORT= 4012
    RUNNING_MESSAGE= 'imrunning'
    CYCLE_PERIOD= 0.01

    def __init__(self, timeModule=time):
        logging.basicConfig(level=logging.DEBUG,
                            format='%(asctime)s %(message)s')
        self._zmq= ZmqRemoteProcedureCall()
        self._replySocket= self._zmq.replySocket(self.REPLY_PORT)
        self._publisherSocket= self._zmq.publisherSocket(self.PUBLISHER_PORT)
        self._counter= 0
        self._timeMod= timeModule
        self._isTerminated= False
        self._logger= Logger.of('MyPublisher')
        self.loop()


    def loop(self):
        self._logger.notice("%s" % self.RUNNING_MESSAGE)
        while not self._isTerminated:
            self._logger.notice('looping counter %d' % self._counter)
            self._zmq.handleRequest(self, self._replySocket,
                                    multi=False)
            anObject= MyObject(self._counter)
            self._zmq.publishPickable(self._publisherSocket, anObject)
            self._counter+= 1
            self._timeMod.sleep(self.CYCLE_PERIOD)
        self._logger.notice("Terminating")


    def getLastCounter(self):
        return self._counter


    @staticmethod
    def startUp():
        MyPublisher()


    @staticmethod
    def commandToSpawn():
        cmd= 'python -c "from %s import MyPublisher; '\
             'MyPublisher.startUp()"' % MyServer.__module__
        return cmd



class MyServer():

    SUCCESS= 123
    REPLY_PORT= 4010
    RUNNING_MESSAGE= 'imrunning'
    TIME_TO_EXECUTE= 2

    def __init__(self, timeModule=time):
        logging.basicConfig(level=logging.DEBUG,
                            format='%(asctime)s %(message)s')
        self._zmq= ZmqRemoteProcedureCall()
        self._replySocket= self._zmq.replySocket(self.REPLY_PORT)
        self._timeMod= timeModule
        self._isTerminated= False
        self._logger= Logger.of('MyServer')
        self.loop()


    def loop(self):
        self._logger.notice("%s" % self.RUNNING_MESSAGE)
        while not self._isTerminated:
            self._logger.notice('looping')
            self._zmq.handleRequest(self, self._replySocket, multi=False)
            self._timeMod.sleep(0.1)
        self._logger.notice("Terminating")


    def returnSuccessAfter2Seconds(self):
        self._logger.notice('invoked returnSuccessAfter2Seconds')
        self._timeMod.sleep(self.TIME_TO_EXECUTE)
        return self.SUCCESS


    def raiseAnException(self):
        self._logger.notice('invoked raiseAnException')
        raise MyException('wanted Exception')


    def getSomethingWithALogger(self, value):
        return HasALogger(value)


    @staticmethod
    def startUpServer():
        MyServer()


    @staticmethod
    def commandToSpawn():
        cmd= 'python -c "from %s import MyServer; '\
             'MyServer.startUpServer()"' % MyServer.__module__
        return cmd


class MyClient():

    def __init__(self):
        self._zmq= ZmqRemoteProcedureCall()
        self._requestSocket= self._zmq.requestSocket(
            'localhost', MyServer.REPLY_PORT)
        self._subscriberSocket= self._zmq.subscriberSocket(
            'localhost', MyPublisher.PUBLISHER_PORT, conflate=True)
        self._requestSocketOnPublisher= self._zmq.requestSocket(
            'localhost', MyPublisher.REPLY_PORT)
        self._logger= Logger.of('MyClient')


    def callASuccessfulMethod(self, timeout):
        return self._zmq.sendRequest(
            self._requestSocket,
            'returnSuccessAfter2Seconds',
            timeout=timeout)


    def callAMethodThatRaisesMyException(self):
        return self._zmq.sendRequest(
            self._requestSocket, 'raiseAnException',
            timeout=2 * MyServer.TIME_TO_EXECUTE)


    def getAnObjectThatHasALogger(self, value):
        return self._zmq.sendRequest(
            self._requestSocket, 'getSomethingWithALogger',
            [value], timeout=1)


    def readFromPublisher(self):
        timeoutInSec= 3.
        obj= self._zmq.receivePickable(self._subscriberSocket,
                                       timeoutInSec)
        self._logger.notice('read %d from publisher' % obj.getValue())
        return obj


    def getLastCounterFromPublisher(self):
        value= self._zmq.sendRequest(
            self._requestSocketOnPublisher, 'getLastCounter',
            [], timeout=1)
        self._logger.notice('asked to publisher and got %d' % value)
        return value


class ZmqRemoteProcedureCallTest(unittest.TestCase):

    THIS_DIR= os.path.abspath(os.path.dirname(__file__))
    TEST_DIR= os.path.join(THIS_DIR, "./tmp/")
    LOG_DIR= os.path.join(TEST_DIR, "log")
    SERVER_LOG_PATH= os.path.join(LOG_DIR, "server.log")
    PUBLISHER_LOG_PATH= os.path.join(LOG_DIR, "publisher.log")



    def setUp(self, timeModule=time):
        logging.basicConfig(level=logging.DEBUG,
                            format='%(asctime)s %(message)s')
        self._removeTestFolderIfItExists()
        self._makeTestDir()
        self._server= None
        self._publisher= None
        self._client= MyClient()
        self._wasSuccessful= False
        self._timeMod= timeModule


    def _makeTestDir(self):
        os.mkdir(self.TEST_DIR)
        os.mkdir(self.LOG_DIR)


    def _removeTestFolderIfItExists(self):
        if os.path.exists(self.TEST_DIR):
            shutil.rmtree(self.TEST_DIR)


    def _spawnProcess(self,
                      commandToSpawn,
                      logFilePath,
                      runningMessage):
        processLog= io.open(os.path.join(logFilePath), "w")
        cmd= "%s" % commandToSpawn()
        print(cmd)
        process= subprocess.Popen(
            shlex.split(cmd),
            stdout=processLog, stderr=processLog
        )
        Poller(5).check(MessageInFileProbe(runningMessage,
                                           logFilePath))
        return process


    def _spawnServer(self):
        self._server= self._spawnProcess(MyServer.commandToSpawn,
                                         self.SERVER_LOG_PATH,
                                         MyServer.RUNNING_MESSAGE)


    def _spawnPublisher(self):
        self._publisher= self._spawnProcess(MyPublisher.commandToSpawn,
                                            self.PUBLISHER_LOG_PATH,
                                            MyPublisher.RUNNING_MESSAGE)



    def tearDown(self):
        print('tearingDown')
        TestHelper.dumpFileToStdout(self.SERVER_LOG_PATH)
        TestHelper.dumpFileToStdout(self.PUBLISHER_LOG_PATH)

        if self._server is not None:
            TestHelper.terminateSubprocess(self._server)
        if self._publisher is not None:
            TestHelper.terminateSubprocess(self._publisher)

        if self._wasSuccessful:
            self._removeTestFolderIfItExists()


    def _testExecuteWithSuccess(self):
        timeoutInSec= 2 * MyServer.TIME_TO_EXECUTE
        ret= self._client.callASuccessfulMethod(timeoutInSec)
        self.assertEqual(
            ret, MyServer.SUCCESS,
            'got %s instead of %s' % (str(ret), MyServer.SUCCESS))


    def _testExecuteAndTimeoutTwiceToAssesSocketIsStillValid(self):
        shortTimeout= 0.5 * MyServer.TIME_TO_EXECUTE
        self.assertRaises(
            ZmqRpcTimeoutError,
            self._client.callASuccessfulMethod, shortTimeout)
        self.assertRaises(
            ZmqRpcTimeoutError,
            self._client.callASuccessfulMethod, shortTimeout)


    def _raiseOnClientSideIfServerRaises(self):
        self.assertRaises(
            MyException,
            self._client.callAMethodThatRaisesMyException)


    def _testReturnObjectWithALogger(self):
        res= self._client.getAnObjectThatHasALogger(13)
        self.assertTrue(res.logger is not None)
        self.assertEqual(13, res.value)


    def _testPublisher(self):
        a= self._client.readFromPublisher()
        b= self._client.readFromPublisher()
        self.assertGreaterEqual(b.getValue(), a.getValue())



    def testMain(self):
        self._spawnServer()
        self._spawnPublisher()
        self._testExecuteWithSuccess()
        self._testExecuteAndTimeoutTwiceToAssesSocketIsStillValid()
        self._raiseOnClientSideIfServerRaises()
        self._testReturnObjectWithALogger()
        self._testPublisher()
        self._wasSuccessful= True


if __name__ == "__main__":
    unittest.main()



import os
import appdirs
from pkg_resources import Requirement, resource_filename
from plico.utils.constants import Constants
from plico.utils.addtree import addtree, mkdirp




__version__= "$Id: calibration_installer.py 50 2018-04-26 21:42:20Z lbusoni $"



class CalibrationInstaller():

    def __init__(self, appDirs, packageName, src=None, dest=None):
        if src is None:
            self._sourcePath= self._getCalibPathInPackage(packageName)
        else:
            self._sourcePath= src
        if dest is None:
            self._destPath= self._getDestinationPathFromAppDirs(appDirs)
        else:
            self._destPath= dest


    def _getDestinationPathFromAppDirs(self, appDirs):
        calibFolderBaseName= 'calib'
        return os.path.join(appDirs.user_data_dir,
                            calibFolderBaseName)


    def getCalibrationFolderPath(self):
        return self._destPath


    def doesCalibrationFolderExists(self):
        return os.path.isdir(self.getCalibrationFolderPath())


    def _createCalibrationFolder(self):
        mkdirp(os.path.dirname(self.getCalibrationFolderPath()))


    def _getCalibPathInPackage(self, packageName):
        return resource_filename(Requirement(packageName),
                                 "%s/calib" % packageName)


    def installCalibrationFilesFromPackage(self):
        self._createCalibrationFolder()
        dest= self.getCalibrationFolderPath()
        addtree(self._sourcePath, dest)



from plico.rpc.abstract_remote_procedure_call import \
    AbstractRemoteProcedureCall
from plico.utils.timeout import Timeout


__version__= "$Id: hackerable_client.py 28 2018-01-27 08:54:00Z lbusoni $"


class HackerableClient(object):


    def __init__(self,
                 rpcHandler,
                 requestSocket,
                 logger):
        assert isinstance(rpcHandler, AbstractRemoteProcedureCall)

        self._rpcHandler= rpcHandler
        self._requestSocket= requestSocket
        self._logger= logger



    def eval(self, expression, timeoutInSec=Timeout.MIRROR_EVAL):
        self._logger.warn("Eval is considered evil")
        return self._rpcHandler.sendRequest(
            self._requestSocket, 'eval',
            [expression], timeout=timeoutInSec)


    def execute(self,
                statement,
                timeoutInSec=Timeout.MIRROR_EXECUTE):
        self._logger.warn("Execute is for commissioning only")
        return self._rpcHandler.sendRequest(
            self._requestSocket, 'execute',
            [statement], timeout=timeoutInSec)




import datetime

__version__= "$Id: timestamp.py 25 2018-01-26 19:00:40Z lbusoni $"


class Timestamp(object):

    def __init__(self):
        self._now= datetime.datetime.now()


    def asNowString(self):
        return self._now.strftime("%Y%m%d_%H%M%S")


    def asTodayString(self):
        return self._now.strftime("%Y%m%d")


    @staticmethod
    def now():
        return Timestamp().asNowString()


    @staticmethod
    def today():
        return Timestamp().asTodayString()


    @staticmethod
    def nowUSec():
        ss= datetime.datetime.now()
        return ss.strftime('%Y%m%d_%H%M%S')


    def __str__(self):
        return self.asNowString()



#!/usr/bin/env python

import random
import threading
import time
import unittest

from test.test_helper import TestHelper, ExecutionProbe, Poller
from plico.utils.decorator import synchronized, override
from plico.utils.fake_convergeable import FakeConvergeable
from plico.utils.concurrent_loop import ConcurrentLoopException, ConcurrentLoop


class FakeOperatorLog(object):

    def __init__(self):
        self._logCnt = 0
        self._mtx = threading.Lock()

    @synchronized("_mtx")
    @override
    def onLoopFailure(self, message):
        self._logCnt += 1

    @synchronized("_mtx")
    def getLoopFailureLogCount(self):
        return self._logCnt


class Test(unittest.TestCase):

    def setUp(self):
        self.interStepTimeSec = 0.001
        self.convergeable = FakeConvergeable()
        self.operatorLog = FakeOperatorLog()
        self.loop = None
        self.buildInitializedLoop()

    def buildUninitializedLoop(self):
        if self.loop is not None:
            self.loop.deinitialize()

        self.loop = ConcurrentLoop(
            "object under test",
            self.convergeable, self.interStepTimeSec,
            self.operatorLog.onLoopFailure)

    def buildInitializedLoop(self):
        self.buildUninitializedLoop()
        self.loop.initialize()

    def tearDown(self):
        self.loop.deinitialize()

    def test_creation(self):
        self.assertFalse(self.loop.isClosed())
        self.assertEqual(0, self.loop.getConvergenceStepCount())
        Poller(2).check(ExecutionProbe(
            lambda: self.assertTrue(
                self.convergeable.getMeasureConvergenceCount() > 1)))

    def test_awakes_from_long_running_sleeps(self):
        self.interStepTimeSec = 10
        cnt = self.convergeable.getMeasureConvergenceCount()
        self.buildInitializedLoop()
        t0Sec = time.time()
        Poller(2).check(ExecutionProbe(
            lambda: self.assertTrue(
                self.convergeable.getMeasureConvergenceCount() > cnt)))
        self.loop.deinitialize()
        t1Sec = time.time()
        self.assertTrue(t1Sec - t0Sec < self.interStepTimeSec / 10)

    def test_close_request_starts_truth_sensor(self):
        self.loop.close()
        self.waitForAtLeastConvergenceSteps(3)
        self.assertTrue(self.loop.isClosed())

    def waitForAtLeastConvergenceSteps(self,
                                       expectedSteps,
                                       timeoutSec=2):
        Poller(timeoutSec).check(ExecutionProbe(
            lambda: self.assertTrue(
                self.convergeable.getConvergenceStepCount() >= expectedSteps)))

    def test_loop_waits_between_steps(self):
        self.interStepTimeSec = 0.02
        self.buildInitializedLoop()
        self.assertEqual(0, self.convergeable.getConvergenceStepCount())
        t0Sec = time.time()
        self.loop.close()
        self.waitForAtLeastConvergenceSteps(2)
        t1Sec = time.time()
        diffSec = t1Sec - t0Sec
        self.assertTrue(t1Sec - t0Sec >= self.interStepTimeSec,
                        ("A loop step must take at least %.3f s but "
                         "it has taken only %.3f s") % (self.interStepTimeSec,
                                                        diffSec))

    def test_waits_between_sensor_failures(self):
        self.interStepTimeSec = 0.1
        self.buildUninitializedLoop()
        self.convergeable.requestPanic()
        self.loop.initialize()
        t0Sec = time.time()
        self.loop.close()
        self.waitForAtLeastConvergenceSteps(2)
        t1Sec = time.time()
        diffSec = t1Sec - t0Sec
        self.assertTrue(diffSec >= self.interStepTimeSec)

    def test_open_request_stops_loop_activity(self):
        self.loop.close()
        self.loop.open()
        cnt = self.convergeable.getConvergenceStepCount()
        self.assertEqual(cnt, self.convergeable.getConvergenceStepCount())
        self.assertEqual(self.loop.getConvergenceStepCount(),
                         self.convergeable.getConvergenceStepCount())
        self.assertFalse(self.loop.isClosed())

    def test_ignores_closing_of_loop_that_is_already_started(self):
        self.loop.close()
        self.loop.close()

    def closeLoopAndWaitForStep(self):
        self.loop.close()
        Poller(2).check(ExecutionProbe(
            lambda: self.assertTrue(
                self.convergeable.getConvergenceStepCount() > 0)))

    def test_complains_when_worker_fails_to_stop(self):
        self.loop.setStopDurationLimitSec(0.001)
        self.convergeable.blockStep()
        self.closeLoopAndWaitForStep()
        self.assertRaises(ConcurrentLoopException,
                          self.loop.deinitialize)
        self.convergeable.unblockStep()

    def test_ignores_open_request_on_open_loop(self):
        self.assertFalse(self.loop.isClosed())
        self.loop.open()

    def test_ignores_step_failures(self):
        self.convergeable.requestPanic()
        self.closeLoopAndWaitForStep()
        Poller(1).check(ExecutionProbe(
            lambda: self.assertTrue(
                self.operatorLog.getLoopFailureLogCount() >= 1)))
        self.loop.open()

    def test_performs_one_pass(self):
        self.loop.performOnePass()
        self.assertEqual(1, self.loop.getConvergenceStepCount())

    def test_convergence_during_close_loop(self):
        self.loop.close()
        self.convergeable.setAsConverged()
        self.assertTrue(self.loop.hasConverged())

        self.convergeable.setAsUnconverged()
        self.assertFalse(self.loop.hasConverged())

    def test_deinitialize_and_reinitialize_restarts_in_open_loop(self):
        self.closeLoopAndWaitForStep()
        self.loop.deinitialize()
        truthSensorClosedLoopCnt = \
            self.convergeable.getConvergenceStepCount()
        truthSensorOpenLoopCnt = \
            self.convergeable.getMeasureConvergenceCount()

        self.loop.initialize()
        self.assertFalse(self.loop.isClosed())
        Poller(4).check(ExecutionProbe(
            lambda: self.assertEqual(
                truthSensorClosedLoopCnt,
                self.convergeable.getConvergenceStepCount())))

        Poller(4).check(ExecutionProbe(
            lambda: self.assertTrue(
                self.convergeable.getConvergenceStepCount() >
                truthSensorOpenLoopCnt)))

        self.loop.close()

        Poller(4).check(ExecutionProbe(
            lambda: self.assertTrue(
                truthSensorClosedLoopCnt <
                self.convergeable.getConvergenceStepCount())))

    @TestHelper.longRunningTest
    def test_concurrent_stress(self):
        N_THREADS = 50
        N_ACTIONS_PER_THREAD = 10
        self.convergeable.setStepSleepDurationSec(0.0005)

        class Stresser(threading.Thread):

            def __init__(self, loop, nIterations):
                threading.Thread.__init__(self)
                self._loop = loop
                self._nIterations = nIterations

            @override
            def run(self):
                for _ in range(0, self._nIterations):
                    try:
                        self._performLoopAction()
                    except Exception:
                        pass

            def _performLoopAction(self):
                N_ACTIONS = 5
                r = random.randint(0, N_ACTIONS - 1)
                if r == 0:
                    self._loop.close()
                elif r == 1:
                    self._loop.open()
                elif r == 2:
                    self._loop.setGain(7.0)
                elif r == 3:
                    self._loop.getGain()
                elif r == 4:
                    self._loop.isClosed()
                else:
                    raise ValueError("Programming mistake")

        stressers = []
        for _ in range(0, N_THREADS):
            stresser = Stresser(self.loop, N_ACTIONS_PER_THREAD)
            stressers.append(stresser)

        for each in stressers:
            each.start()

        for each in stressers:
            each.join(10)
            assert not each.isAlive()


if __name__ == "__main__":
    unittest.main()




import abc
import time
import traceback
from six import with_metaclass

__version__= "$Id: control_loop.py 25 2018-01-26 19:00:40Z lbusoni $"


class ControlLoop(with_metaclass(abc.ABCMeta, object)):
    @abc.abstractmethod
    def start(self):
        assert False, "Abstract method"


class FaultTolerantControlLoop(ControlLoop):

    def __init__(self,
                 steppable,
                 logger,
                 timeModule=time,
                 loopPeriodInSecond=1):
        self._steppable= steppable
        self._logger= logger
        self._timeModule= timeModule
        self._loopPeriodSec= loopPeriodInSecond


    def start(self):
        while self._isAlive():
            try:
                self._steppable.step()
            except Exception as e:
                traceback.print_exc()
                self._logger.error(str(e))
            self._timeModule.sleep(self._loopPeriodSec)


    def _isAlive(self):
        return not self._steppable.isTerminated()



#!/usr/bin/env python
import unittest
import os
from plico.utils.configuration import Configuration

__version__ = "$Id: configuration_test.py 26 2018-01-26 19:06:25Z lbusoni $"


class ConfigurationTest(unittest.TestCase):


    def setUp(self):
        self._hackToRunTestSuiteFromTrunkDir()

    def _hackToRunTestSuiteFromTrunkDir(self):
        print((os.getcwd()))
        if os.path.basename(os.getcwd()) == 'utils':
            self.CONF_FILE= 'configExample.conf'
        else:
            self.CONF_FILE= 'test/utils/configExample.conf'


    def testGetEntry(self):
        cfg= Configuration()
        cfg.load(self.CONF_FILE)
        self.assertEqual('192.168.1.29',
                         cfg.getValue('camera', 'ip_address'))


    def testModelIsReported(self):
        cfg= Configuration()
        cfg.load(self.CONF_FILE)
        self.assertEqual('bar', cfg.deviceModel('foo'))


    def testRaisesIfConfigFileIsMissing(self):
        cfg= Configuration()
        self.assertRaises(Exception, cfg.load, 'foo')


    def testGetFloatValue(self):
        cfg= Configuration()
        cfg.load(self.CONF_FILE)
        self.assertEqual(
            1.23e-9,
            cfg.getValue('foo', 'a_float_value', getfloat=True))


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python

import os
import configparser
import appdirs
from pkg_resources import resource_filename
from plico.utils.calibration_installer import CalibrationInstaller
from plico.utils.addtree import mkdirp


__version__= "$Id: configuration.py 50 2018-04-26 21:42:20Z lbusoni $"


class LogFolderManager():

    def __init__(self, appDirs):
        self._appdirs= appDirs


    def getLogFolderPath(self):
        return self._appdirs.user_log_dir


    def doesLogFolderExists(self):
        return os.path.isdir(self.getLogFolderPath())


    def createLogFolder(self):
        mkdirp(os.path.dirname(self.getLogFolderPath()))


class Configuration(object):


    def __init__(self):
        NOT_INITIALIZED= None
        self._cfg = configparser.ConfigParser()


    def networkHostName(self, section):
        return self.getValue(section, 'host')


    def calibrationRootDir(self):
        return self._getCalibDir()


    def loggingDir(self):
        return self._getLogDir()


    def deviceModel(self, section):
        return self.getValue(section, 'model')


    def deviceName(self, section):
        return self.getValue(section, 'name')


    def basePort(self, section):
        return self.getValue(section, 'port', getint=True)


    def logLevel(self, section):
        return self.getValue(section, 'log_level')


    def _assertSectionExists(self, section):
        if section not in self._cfg.sections():
            raise Exception('Section "%s" not found in configuration file' %
                            section)


    def getValue(self, section, entry, getint=False,
                 getfloat=False, getboolean=False):
        if entry not in list(dict(self._cfg.items(section)).keys()):
            raise KeyError(
                'Section "%s" must contain a "%s" entry. File is %s. Keys %s' %
                (section, entry, self._filename,
                 list(dict(self._cfg.items(section)).keys())))
        if getint:
            try:
                value = self._cfg.getint(section, entry)
            except ValueError:
                raise ValueError(
                    'Value for entry "%s" in section "%s"'
                    ' must be an integer number' % (entry, section))
        elif getfloat:
            try:
                value = self._cfg.getfloat(section, entry)
            except ValueError:
                raise ValueError(
                    'Value for entry "%s" in section "%s"'
                    ' must be a float number' % (entry, section))
        elif getboolean:
            try:
                value = self._cfg.getboolean(section, entry)
            except ValueError:
                raise ValueError(
                    'Value for entry "%s" in section "%s"'
                    ' must be a boolean' % (entry, section))
        else:
            value = self._cfg.get(section, entry)
        return value


    def load(self, filename):
        if not os.path.exists(filename):
            raise Exception('Configuration file not found: %s' % filename)
        self._cfg.read(filename)
        self._filename= filename


    def _getLogDir(self):
        try:
            return self.getValue('global', 'force_log_dir')
        except Exception:
            lfm= LogFolderManager(self._getAppDirs())
            lfm.createLogFolder()
            return lfm.getLogFolderPath()


    def _getCalibDir(self):
        try:
            return self.getValue('global', 'force_calib_folder_dest')
        except Exception:
            calibInst= CalibrationInstaller(
                self._getAppDirs(), self._getPythonPackageName())
            calibInst.installCalibrationFilesFromPackage()
            return calibInst.getCalibrationFolderPath()


    def _getPythonPackageName(self):
            return self.getValue('global', 'python_package_name')


    def _getAppName(self):
            return self.getValue('global', 'app_name')


    def _getAppAuthor(self):
            return self.getValue('global', 'app_author')


    def _getAppDirs(self):
        return appdirs.AppDirs(self._getAppName(),
                               self._getAppAuthor())


from plico.rpc.abstract_remote_procedure_call import AbstractRemoteProcedureCall
from plico.utils.decorator import override


class DummyRpcHandler(AbstractRemoteProcedureCall):

    def __init__(self):
        pass


    @override
    def sendRequest(self):
        pass


    @override
    def handleRequest(self):
        pass


    @override
    def recvCameraFrame(self):
        pass


    @override
    def sendCameraFrame(self):
        pass


    @override
    def publishPickable(self):
        pass


    @override
    def receivePickable(self):
        pass



#!/usr/bin/env python
import unittest
from plico.rpc.zmq_ports import ZmqPorts



class ZmqPortsTest(unittest.TestCase):


    def setUp(self):
        self._zmqPorts= ZmqPorts(None, None)



    def testHostname(self):
        self.assertTrue(True)


if __name__ == "__main__":
    unittest.main()





import threading
from plico.utils.logger import Logger
from plico.utils.decorator import override, synchronized, returns
from plico.utils.loop import Loop

__version__ = "$Id: concurrent_loop.py 25 2018-01-26 19:00:40Z lbusoni $"


class WorkerThread(threading.Thread):

    def __init__(self, name,
                 closedFunc,
                 convergable,
                 interStepTimeSec,
                 stepCounter,
                 logFailureFunc):
        threading.Thread.__init__(self)
        self._name = name
        self._closedFunc = closedFunc
        self._convergable = convergable
        self._stopDurationLimitSec = 5
        self._stepCounter = stepCounter
        self._logFailureFunc = logFailureFunc
        self._interStepTimeSec = interStepTimeSec

        self._runEnabled = True
        self._runEnabledCondition = threading.Condition()
        self._doneEvent = threading.Event()
        self._logger = Logger.of("ConcurrentLoop.WorkerThread")

    @override
    def run(self):
        while self._isRunEnabled():
            try:
                self._step()
            except Exception as e:
                # import traceback; traceback.print_exc()
                self._logFailureFunc(str(e))
            self._waitForNextStep()
        self._flagAsDone()

    @synchronized("_runEnabledCondition")
    def _waitForNextStep(self):
        if self._runEnabled:
            self._runEnabledCondition.wait(self._interStepTimeSec)

    def _markAsClosed(self):
        self._openEvent.clear()
        self._closeEvent.set()

    def _markAsOpen(self):
        self._openEvent.set()
        self._closeEvent.clear()

    def _step(self):
        if self._closedFunc():
            self._stepCounter.inc()
            self._convergable.performOneConvergenceStep()
        else:
            self._convergable.measureConvergence()

    def _flagAsDone(self):
        self._doneEvent.set()
        assert self._doneEvent.isSet()

    def _isDone(self):
        return self._doneEvent.isSet()

    @synchronized("_runEnabledCondition")
    def _isRunEnabled(self):
        return self._runEnabled

    @synchronized("_runEnabledCondition")
    def _disableRun(self):
        self._runEnabled = False
        self._runEnabledCondition.notifyAll()

    def setStopDurationLimitSec(self, limitSec):
        self._stopDurationLimitSec = limitSec

    def _waitForDone(self):
        self._doneEvent.wait(self._stopDurationLimitSec)
        if not self._isDone():
            raise ConcurrentLoopException(
                "Failed to stop %s within limit of %.3f s" % (
                    self._name, self._stopDurationLimitSec))
        assert self._isDone()

    def stop(self):
        self._logger.notice("Stopping ...")
        self._disableRun()
        self._waitForDone()
        self._logger.notice("Stopped.")


class ConcurrentLoopException(Exception):
    pass


class ThreadSafeStepCounter(object):

    def __init__(self):
        self._cnt = 0
        self._mtx = threading.Lock()

    @synchronized("_mtx")
    def inc(self):
        self._cnt += 1

    @synchronized("_mtx")
    def getCount(self):
        return self._cnt


class ConcurrentLoop(Loop):

    def __init__(self, name, convergable, interStepTimeSec, logFailureFunc):
        Loop.__init__(self)
        self._name = name
        self._convergable = convergable
        self._interStepTimeSec = interStepTimeSec
        self._logFailureFunc = logFailureFunc

        self._closed = False
        self._closedMtx = threading.Lock()
        self._logger = Logger.of("ConcurrentLoop")
        self._mtx = threading.Lock()
        self._stopDurationLimitSec = 5
        self._worker = None
        self._stepCounter = ThreadSafeStepCounter()
        self._initialized = False

    @synchronized("_mtx")
    def isInitialized(self):
        return self._initialized

    def initialize(self):
        if not self.isInitialized():
            self._startWorkerThread()
            with self._mtx:
                self._initialized = True

    def _startWorkerThread(self):
        assert self._worker is None
        self._worker = WorkerThread("Worker thread for %s" % self.name(),
                                   self.isClosed, self._convergable,
                                   self._interStepTimeSec,
                                   self._stepCounter,
                                   self._logFailureFunc)
        self._worker.setStopDurationLimitSec(self._stopDurationLimitSec)
        self._worker.start()
        self._logger.notice("Worker thread is started.")

    @override
    def name(self):
        return self._name

    @override
    def performOnePass(self):
        if self.isInitialized():
            if not self.isClosed():
                self._stepCounter.inc()
                self._convergable.performOneConvergenceStep()
            else:
                raise Exception("Loop is closed")
        else:
            self._logger.error("%s is not yet initialized" % self._name)
#             raise Exception("%s is not yet initialized" % self._name)

    @synchronized("_mtx")
    @override
    def close(self):
        if self._initialized:
            self._flagAsClosed()
        else:
            self._logger.error("%s is not yet initialized" % self._name)
#             raise Exception("%s is not yet initialized" % self._name)

    @synchronized("_mtx")
    def setStopDurationLimitSec(self, durationSec):
        self._stopDurationLimitSec = durationSec
        if self._worker:
            self._worker.setStopDurationLimitSec(durationSec)

    @synchronized("_mtx")
    @override
    def open(self):
        if self._initialized:
            self._flagAsOpen()
        else:
            self._logger.error("%s is not yet initialized" % self._name)

    @synchronized("_closedMtx")
    def _flagAsClosed(self):
        self._closed = True

    @synchronized("_closedMtx")
    def _flagAsOpen(self):
        self._closed = False

    @synchronized("_closedMtx")
    @override
    @returns(bool)
    def isClosed(self):
        return self._closed

    @override
    @returns(int)
    def getConvergenceStepCount(self):
        return self._stepCounter.getCount()

    @override
    def hasConverged(self):
        return self._convergable.hasConverged()

    def _stopWorkerThread(self):
        self._worker.stop()
        self._worker.join(self._stopDurationLimitSec)
        assert not self._worker.is_alive()
        self._worker = None

    def deinitialize(self):
        self._logger.debug("Deinitializing ...")
        if self._worker is not None:
            self._stopWorkerThread()
        with self._mtx:
            self._initialized = False
            self._closed = False
            self._worker = None
            self._logger.notice("Loop is deinitialized.")



from plico.utils.constants import Constants

__version__= "$Id: zmq_ports.py 56 2018-09-14 16:42:15Z lbusoni $"



class ZmqPorts(object):

    def __init__(self, hostname, basePort):
        self._hostname= hostname
        self._basePort= basePort


    @staticmethod
    def fromConfiguration(configuration, configSection):
        hostname= configuration.networkHostName(configSection)
        basePort= configuration.basePort(configSection)
        return ZmqPorts(hostname, basePort)


    @property
    def SERVER_REPLY_PORT(self):
        return self._basePort + Constants.PORT_REPLY_OFFSET


    @property
    def SERVER_STATUS_PORT(self):
        return self._basePort + Constants.PORT_STATUS_OFFSET


    @property
    def SERVER_PUBLISHER_PORT(self):
        return self._basePort + Constants.PORT_PUBLISHER_OFFSET


    @property
    def SERVER_DISPLAY_PORT(self):
        return self._basePort + Constants.PORT_DISPLAY_OFFSET


    @property
    def SERVER_HOSTNAME(self):
        return self._hostname



'''
Created on Nov 5, 2019

@author: lbusoni
'''

import logging


def setUpLogger(filePath, loggingLevel):
    import logging.handlers
    FORMAT = '%(asctime)s %(levelname)s %(name)s %(message)s'
    f = logging.Formatter(fmt=FORMAT)
    handler = logging.handlers.RotatingFileHandler(
        filePath, encoding='utf8', maxBytes=1000, backupCount=3)
    root_logger = logging.getLogger()
    root_logger.setLevel(loggingLevel)
    handler.setFormatter(f)
    handler.setLevel(loggingLevel)
    root_logger.addHandler(handler)
    handler.doRollover()


def pippo(a):
    logger = logging.getLogger('IlLoggerDiPippo')
    logger.debug('debug mi hai passato %d' % a)
    if a < 0:
        logger.error('error a<0 (a=%g)' % a)


class Pluto():

    def __init__(self):
        self._logger = logging.getLogger('PLUTO')

    def funz1(self, a):
        self._logger.debug('debug mi hai passato %d' % a)
        if a < 0:
            self._logger.error('error a<0 (a=%g)' % a)




import abc
from plico.utils.decorator import returns
from six import with_metaclass

__version__= "$Id: convergeable.py 25 2018-01-26 19:00:40Z lbusoni $"


class Convergeable(with_metaclass(abc.ABCMeta, object)):

    @abc.abstractmethod
    @returns(bool)
    def hasConverged(self):
        assert False


    @abc.abstractmethod
    def performOneConvergenceStep(self):
        assert False


    @abc.abstractmethod
    def measureConvergence(self):
        assert False



import logging
import os
import datetime


__version__= "$Id: starter_script_creator.py 31 2018-01-27 10:47:29Z lbusoni $"


class StarterScriptCreatorBase(object):

    def __init__(self):
        NOT_INITIALIZED=None
        self._binDir= NOT_INITIALIZED
        self._pythonPath= NOT_INITIALIZED
        self._configFileDest= NOT_INITIALIZED


    def _now(self):
        return datetime.datetime.now()


    def _createFoldersIfMissing(self, filename):
        if not os.path.isdir(os.path.dirname(filename)):
            os.makedirs(os.path.dirname(filename))


    def setInstallationBinDir(self, binDir):
        self._binDir= binDir


    def setPythonPath(self, pythonPath):
        self._pythonPath= pythonPath


    def setConfigFileDestination(self, configFileDest):
        self._configFileDest= configFileDest


    def _createAStarterScript(self,
                              destinationPath,
                              executableSrc,
                              configFileSection,
                              command='python'):
        logging.info("Creating starter script %s" % destinationPath)
        self._createFoldersIfMissing(destinationPath)
        with open(destinationPath, "w") as text_file:
            text_file.write("#!/bin/bash\n")
            text_file.write("# Auto-generated on %s - %s\n\n" % (
                str(self._now()), __version__))
            text_file.write('export PYTHONPATH="%s":$PYTHONPATH\n' %
                            self._pythonPath)
            text_file.write("%s '%s' '%s' %s\n" % (
                command,
                os.path.join(self._pythonPath, executableSrc),
                self._configFileDest,
                configFileSection))

        self._setExecutableFlag(destinationPath)


    def _setExecutableFlag(self, filePath):
        self._execute('chmod +x "%s"' % filePath)


    def _execute(self, cmd):
        logging.info(cmd)
        exitCode= os.system(cmd)
        assert os.EX_OK == exitCode, "Command '%s' must succeed" % cmd


    def installExecutables(self):
        assert False, 'method to be implemented by derived classes'




__version__ = "$Id: constants.py 56 2018-09-14 16:42:15Z lbusoni $"


class Constants:
    PICKLE_PROTOCOL= 2
    APP_NAME= "inaf.arcetri.ao.plico"
    APP_AUTHOR= "INAF Arcetri Adaptive Optics"

    PORT_REPLY_OFFSET= 0
    PORT_STATUS_OFFSET= 1
    PORT_PUBLISHER_OFFSET= 2
    PORT_DISPLAY_OFFSET= 3



import abc
from six import with_metaclass

__version__= "$Id: loop.py 25 2018-01-26 19:00:40Z lbusoni $"


class Loop(with_metaclass(abc.ABCMeta, object)):

    @abc.abstractmethod
    def name(self):
        assert False


    @abc.abstractmethod
    def close(self):
        assert False


    @abc.abstractmethod
    def open(self):
        assert False


    @abc.abstractmethod
    def isClosed(self):
        assert False


    @abc.abstractmethod
    def performOnePass(self):
        assert False


    @abc.abstractmethod
    def getConvergenceStepCount(self):
        assert False


    @abc.abstractmethod
    def hasConverged(self):
        assert False



class LoopException(Exception):

    def __init__(self, message):
        Exception.__init__(self, message)



from __future__ import print_function

import logging
import logging.handlers
import os
import sys
import signal
from plico.utils.configuration import Configuration
from plico.rpc.zmq_remote_procedure_call import ZmqRemoteProcedureCall
from plico.rpc.sockets import Sockets
from plico.rpc.zmq_ports import ZmqPorts

__version__= "$Id: base_runner.py 54 2018-04-30 21:00:26Z lbusoni $"


class BaseRunner(object):

    def __init__(self):
        INITIALIZED_LATER= None
        self._configFilePath= INITIALIZED_LATER
        self._rpc= INITIALIZED_LATER
        self._configuration= INITIALIZED_LATER
        self._sockets= INITIALIZED_LATER
        self._argv= INITIALIZED_LATER


    def _logRunning(self):
        self._logger.notice(self.RUNNING_MESSAGE)
        sys.stdout.flush()


    def _createFoldersIfMissing(self, filename):
        if not os.path.isdir(os.path.dirname(filename)):
            os.makedirs(os.path.dirname(filename))


    def _determineLoggingLevel(self):
        try:
            levelAsStr= self._configuration.logLevel(
                self._configurationSection)
        except KeyError:
            levelAsStr= 'debug'
        if levelAsStr == 'fatal':
            loggingLevel= logging.CRITICAL
        elif levelAsStr == 'error':
            loggingLevel= logging.ERROR
        elif levelAsStr == 'warning':
            loggingLevel= logging.WARNING
        elif levelAsStr == 'info':
            loggingLevel= logging.INFO
        elif levelAsStr == 'debug':
            loggingLevel= logging.DEBUG
        else:
            loggingLevel= logging.DEBUG
        return loggingLevel


    def _setUpLogging(self, logFilePath):
        loggingLevel= self._determineLoggingLevel()
        print(('Setting up logging in file %s level %s' % (
            logFilePath, loggingLevel)))
        self._createFoldersIfMissing(logFilePath)
        FORMAT='%(asctime)s %(levelname)s %(message)s'
        f = logging.Formatter(fmt=FORMAT)
        rotHandler= logging.handlers.RotatingFileHandler(
            logFilePath, encoding='utf8',
            maxBytes=10000000, backupCount=10)
        handlers = [rotHandler]
        root_logger = logging.getLogger()
        root_logger.setLevel(loggingLevel)
        for h in handlers:
            h.setFormatter(f)
            h.setLevel(loggingLevel)
            root_logger.addHandler(h)
        rotHandler.doRollover()


    def _createConfiguration(self):
        self._configuration= Configuration()
        self._configuration.load(self._configFilePath)


    def _createZmqBasedRPC(self):
        zmqPorts= ZmqPorts.fromConfiguration(self._configuration,
                                             self._configurationSection)
        self._rpc= ZmqRemoteProcedureCall()
        self._sockets= Sockets(zmqPorts, self._rpc)


    def _checkCommandLine(self, argv):
        if 3 > len(argv):
            programName= argv[0]
            print("Usage:", file=sys.stderr)
            print("\t%s <configuration_file> <configuration_section>" %
                  programName, file=sys.stderr)
            print("Example:", file=sys.stderr)
            print(("\t%s /home/plico/.plico/plico.conf") %
                  programName, file=sys.stderr)
            print("\nargv was: %s" % str(argv))
            sys.exit(os.EX_USAGE)
        print('argv: %s' % str(argv))


    def _parseCommandLine(self, argv):
        self._checkCommandLine(argv)
        self._configFilePath= argv[1]
        self._configurationSection= argv[2]


    def _logFilePath(self):
        return os.path.join(self._configuration.loggingDir(),
                            '%s.log' % self._configurationSection)


    def start(self, argv):
        self._argv= argv
        self._parseCommandLine(argv)
        self._createConfiguration()
        self._setUpLogging(self._logFilePath())
        self._createZmqBasedRPC()
        self._registrerHandlers()
        return self.run()


    def run(self):
        pass


    def getConfigFilePath(self):
        return self._configFilePath


    @property
    def name(self):
        return self._configuration.getValue(
            self._configurationSection,
            'name')


    @property
    def configuration(self):
        return self._configuration


    def getConfigurationSection(self):
        return self._configurationSection


    def rpc(self):
        return self._rpc


    def sockets(self):
        return self._sockets


    def argv(self):
        return self._argv


    def terminate(self, signal, frame):
        pass


    def _registerHandlerForSigInt(self):
        signal.signal(signal.SIGINT, self.terminate)


    def _registerHandlerForSigTerm(self):
        signal.signal(signal.SIGTERM, self.terminate)


    def _registerHandlerForSigKill(self):
        try:
            signal.signal(signal.SIGKILL, self.terminate)
        except Exception:
            pass


    def _registrerHandlers(self):
        self._registerHandlerForSigInt()
        self._registerHandlerForSigTerm()
        self._registerHandlerForSigKill()



import abc
import time
from plico.utils.decorator import override
from six import with_metaclass


__version__= "$Id: barrier.py 25 2018-01-26 19:00:40Z lbusoni $"


class Predicate(with_metaclass(abc.ABCMeta, object)):

    @abc.abstractmethod
    def isFullfilled(self):
        assert False


    @abc.abstractmethod
    def errorMessage(self):
        assert False


class FunctionPredicate(Predicate):

    def __init__(self, nulladicFunction, functionName=""):
        self._nulladicFunction= nulladicFunction
        self._functionName= functionName


    @staticmethod
    def create(func, *args, **kwds):
        return FunctionPredicate(lambda: func(*args, **kwds),
                                 functionName=func.__name__)


    @override
    def isFullfilled(self):
        return self._nulladicFunction()


    @override
    def errorMessage(self):
        return "Predicate '%s' has failed." % self._functionName


class BarrierTimeout(Exception):

    def __init__(self, message):
        Exception.__init__(self, message)


class Barrier:

    @classmethod
    def waitUntil(cls, predicate, timeoutSec, period=0.5, timeModule=time):
        mustEnd = timeModule.time() + timeoutSec

        done= False
        while not done:
            ok= predicate()
            if ok:
                done= True
            elif timeModule.time() >= mustEnd:
                raise BarrierTimeout("Timeout occurred after %.1f s" %
                                     timeoutSec)
            else:
                timeModule.sleep(period)


    def __init__(self, timeoutSec, pollingPeriodSec=0.1, timeModule=time):
        self._timeoutSec= timeoutSec
        self._pollingPeriodSec= pollingPeriodSec
        self._timeModule= timeModule


    def waitFor(self, predicate):
        try:
            Barrier.waitUntil(
                predicate.isFullfilled, self._timeoutSec,
                self._pollingPeriodSec, self._timeModule)
        except BarrierTimeout:
            raise BarrierTimeout("Timeout occurred after %.1f s: %s" % (
                                 self._timeoutSec, predicate.errorMessage()))



import abc
from plico.utils.decorator import returnsNone
from six import with_metaclass

__version__= "$Id: abstract_remote_procedure_call.py 25 2018-01-26 19:00:40Z lbusoni $"


class AbstractRemoteProcedureCall(with_metaclass(abc.ABCMeta, object)):
    @abc.abstractmethod
    @returnsNone
    def sendRequest(self):
        assert False


    @abc.abstractmethod
    def handleRequest(self):
        assert False


    @abc.abstractmethod
    def recvCameraFrame(self):
        assert False


    @abc.abstractmethod
    @returnsNone
    def sendCameraFrame(self):
        assert False


    @abc.abstractmethod
    @returnsNone
    def publishPickable(self, socket, pickableObject):
        assert False


    @abc.abstractmethod
    def receivePickable(self, socket, timeoutInSec=10):
        assert False



import time


__version__= "$Id: timekeeper.py 25 2018-01-26 19:00:40Z lbusoni $"


class TimeKeeper(object):

    def __init__(self, interval=1.0, timeMod=time):
        self.interval = interval
        self.t0= timeMod.time()
        self.cnt = 0

    def inc(self):
        self.cnt +=1
        t1 = time.time()
        if (t1 - self.t0)>=self.interval:
            tdiff = t1 - self.t0
            self.rate = self.cnt / tdiff
            self.count = self.cnt
            self.ms = 1.0 / self.rate * 1e3
            self.cnt=0
            self.t0=t1
            return True
        return False




from abc import abstractmethod
from functools import wraps
import abc
import threading
import traceback

from concurrent import futures
from plico.utils.decorator import override, synchronized
from plico.utils.logger import Logger
from six import with_metaclass



__version__= "$Id: executor.py 25 2018-01-26 19:00:40Z lbusoni $"


class Task(with_metaclass(abc.ABCMeta, object)):

    @abstractmethod
    def perform(self):
        assert False, "abstract base clase"


    def name(self):
        return "unspecified"


    def __repr__(self):
        return "Task(%r)" % self.name()


class FunctionTaskException(Exception):
    pass


class FunctionTask(Task):

    NO_RESULT= object()

    def __init__(self, func, *args, **kwds):
        self._func= func
        self._args= args
        self._kwds= kwds

        self._result= self.NO_RESULT
        self._exception= None
        self._name= None


    @override
    def perform(self):
        try:
            self._result= self._func(*self._args, **self._kwds)
        except Exception as e:
            self._exception= e
            raise


    @override
    def name(self):
        if self._name:
            return self._name

        return "%s" % self._func


    def __repr__(self):
        return "FunctionTask(%r)" % self.name()


    def withName(self, name):
        self._name= name
        return self


    def getResult(self):
        if self._result is self.NO_RESULT:
            raise FunctionTaskException(
                "No result available for %s" % self.name())
        return self._result


    def getException(self):
        return self._exception


    def reraiseExceptionIfAny(self):
        if self._exception:
            raise self._exception


class TaskControl(with_metaclass(abc.ABCMeta, object)):
    @abstractmethod
    def isDone(self):
        assert False, "abstract base class"


    @abstractmethod
    def isRunning(self):
        assert False, "abstract base class"


    @abstractmethod
    def waitForCompletion(self, timeoutSec):
        assert False, "abstract base class"



class TaskTimeoutError(Exception):

    def __init__(self, msg):
        Exception.__init__(self, msg)


class FutureTaskControl(TaskControl):

    def __init__(self, future, taskName, logger):
        TaskControl.__init__(self)
        assert taskName is not None
        self._future= future
        self._taskName= taskName
        self._logger= logger


    @override
    def isDone(self):
        return self._future.done()


    @override
    def isRunning(self):
        return self._future.running()


    def _printStackTraceInCaseOfError(self):
        try:
            self._future.result()
        except Exception as e:
            self._logger.error("Task '%s' failed: %s" % (
                               self._taskName, str(e)))
            traceback.print_exc()
            self._logger.debug("Details to %s: %s" % (
                               self._taskName, traceback.format_exc()))


    @override
    def waitForCompletion(self, timeoutSec=None):
        doneFutures, _= futures.wait([self._future], timeout=timeoutSec)
        if self._future in doneFutures:
            self._printStackTraceInCaseOfError()
        else:
            raise TaskTimeoutError("Wait for completion"
                                   " of task '%s' failed" % self._taskName)


class Executor(with_metaclass(abc.ABCMeta, object)):

    @abstractmethod
    def execute(self, task):
        assert False, "abstract base class"


class DoneTaskControl(TaskControl):

    @override
    def isDone(self):
        return True


    @override
    def isRunning(self):
        return False


    @override
    def waitForCompletion(self, timeoutSec):
            pass


class SerialExecutor(Executor):

    def __init__(self, logger=Logger.of("SerialExecutor")):
        Executor.__init__(self)
        self._logger= logger


    @override
    def execute(self, task):
        assert task is not None
        self._logger.debug("Executing task '%s'" % task.name())
        try:
            task.perform()
        except Exception as e:
            self._logger.error(str(e))

        return DoneTaskControl()


class MultiThreadExecutor(Executor):

    def __init__(self, maxThreads, logger=Logger.of("MultiThreadExecutor")):
        Executor.__init__(self)
        self._logger= logger
        self._executor= futures.ThreadPoolExecutor(max_workers=maxThreads)


    @override
    def execute(self, task):
        assert task is not None
        self._logger.debug("Executing task '%s'" % task.name())
        future= self._executor.submit(task.perform)
        return FutureTaskControl(future, task.name(), self._logger)


class SingleThreadExecutor(MultiThreadExecutor):

    def __init__(self, logger=Logger.of("SingleThreadExecutor")):
        MultiThreadExecutor.__init__(self, maxThreads=1, logger=logger)


class BusyExecutorException(Exception):

    def __init__(self, message, nameOfCurrentTask):
        Exception.__init__(self, message)
        self.nameOfCurrentTask= nameOfCurrentTask


class SingleThreadImmediateExecutor(Executor):

    def __init__(self, singleThreadExecutor):
        self._executor= singleThreadExecutor
        self._idle= True
        self._mutex= threading.RLock()
        self._currentTask= None


    @override
    @synchronized("_mutex")
    def execute(self, task):
        if not self._idle:
            taskName= self._currentTask.name()
            raise BusyExecutorException(
                "Executor is busy by task '%s'" % (taskName), taskName)

        class TrackedTask(Task):
            def __init__(self, task, onCompletionCallbackFunction):
                Task.__init__(self)
                self._task= task
                self._onCompletionCallback= onCompletionCallbackFunction

            @override
            def perform(self):
                try:
                    self._task.perform()
                finally:
                    self._onCompletionCallback()

            @override
            def name(self):
                return self._task.name()

        self._idle= False
        self._currentTask= task
        return self._executor.execute(TrackedTask(task, self._onCompletion))


    @synchronized("_mutex")
    def _onCompletion(self):
        self._currentTask= None
        self._idle= True


    @synchronized("_mutex")
    def isIdle(self):
        return self._idle


class DelayedTaskControl(TaskControl):


    def __init__(self):
        super(DelayedTaskControl, self).__init__()
        self._done= False


    @override
    def isDone(self):
        return self._done


    def markAsDone(self):
        self._done= True


    @override
    def isRunning(self):
        return False


    @override
    def waitForCompletion(self, timeoutSec):
        assert False, "N/A"


class TaskDelayingExecutor(Executor):


    def __init__(self):
        super(TaskDelayingExecutor, self).__init__()
        self._delayedTasks= []

    @override
    def execute(self, task):
        class WrappedTask(Task):
            def __init__(self, task, delayedTaskControl):
                super(WrappedTask, self).__init__()
                self._task= task
                self._taskControl= delayedTaskControl

            def perform(self):
                self._task.perform()
                self._taskControl.markAsDone()


        taskCtrl= DelayedTaskControl()
        self._delayedTasks.append(WrappedTask(task, taskCtrl))
        return taskCtrl


    def getDelayedTasks(self):
        return self._delayedTasks


    def executeDelayedTasks(self):
        for each in self._delayedTasks:
            each.perform()


class ExecutorFactory(object):


    @staticmethod
    def buildMultiThreadExecutor(maxNumThreads):
        return MultiThreadExecutor(maxNumThreads)


    @staticmethod
    def buildSingleThreadExecutor(logger=Logger.of("SingleThreadExecutor")):
        return SingleThreadExecutor(logger)


    @staticmethod
    def buildSingleThreadImmediateExecutor(
            logger=Logger.of("SingleThreadImmediateExecutor")):
        return SingleThreadImmediateExecutor(
            ExecutorFactory.buildSingleThreadExecutor(logger))



def logTaskFailure(func):
    @wraps(func)
    def wrappedMethod(self, *args, **kwds):
        try:
            return func(self, *args, **kwds)
        except Exception as e:
            message= "Task failed: %s" % str(e)
            self._logger.error(message)
            raise

    return wrappedMethod



#!/usr/bin/env python
import unittest
from plico.utils.hackerable import Hackerable
from plico.utils.logger import Logger

__version__ = "$Id: hackerable_test.py 26 2018-01-26 19:06:25Z lbusoni $"


class MyHackerable(Hackerable):

    class GenericObject(object):

        def __init__(self):
            self._value= 0

        def setValue(self, value):
            self._value= value

        def getValue(self):
            return self._value


    def __init__(self):
        self._aMember= self.GenericObject()
        self._logger= Logger.of('MyLogger')
        Hackerable.__init__(self, self._logger)


class HackerableTest(unittest.TestCase):


    def setUp(self):
        self.hack= MyHackerable()


    def testEvalExecute(self):
        self.hack.execute("self._aMember.setValue(42.0)")
        self.assertEqual("42.0",
                         self.hack.eval("self._aMember.getValue()"))



if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env/python
from plico.utils.decorator import cacheResult


__version__= "$Id: sockets.py 56 2018-09-14 16:42:15Z lbusoni $"



class Sockets():

    def __init__(self, ports, zmqRpc):
        self._zmqRpc= zmqRpc
        self._ports= ports


    @cacheResult
    def serverSubscriber(self):
        return self._zmqRpc.subscriberSocket(
            self._ports.SERVER_HOSTNAME,
            self._ports.SERVER_PUBLISHER_PORT, conflate=True)


    @cacheResult
    def serverSubscriberAddress(self):
        return self._zmqRpc.tcpAddress(
            self._ports.SERVER_HOSTNAME,
            self._ports.SERVER_PUBLISHER_PORT)


    @cacheResult
    def serverRequest(self):
        return self._zmqRpc.requestSocket(
            self._ports.SERVER_HOSTNAME, self._ports.SERVER_REPLY_PORT)


    @cacheResult
    def serverRequestAddress(self):
        return self._zmqRpc.tcpAddress(
            self._ports.SERVER_HOSTNAME,
            self._ports.SERVER_REPLY_PORT)


    @cacheResult
    def serverStatus(self):
        return self._zmqRpc.subscriberSocket(
            self._ports.SERVER_HOSTNAME,
            self._ports.SERVER_STATUS_PORT, conflate=True)


    @cacheResult
    def serverStatusAddress(self):
        return self._zmqRpc.tcpAddress(
            self._ports.SERVER_HOSTNAME,
            self._ports.SERVER_STATUS_PORT)


    @cacheResult
    def serverDisplay(self):
        return self._zmqRpc.subscriberSocket(
            self._ports.SERVER_HOSTNAME,
            self._ports.SERVER_DISPLAY_PORT, conflate=True)


    @cacheResult
    def serverDisplayAddress(self):
        return self._zmqRpc.tcpAddress(
            self._ports.SERVER_HOSTNAME,
            self._ports.SERVER_DISPLAY_PORT)



#!/usr/bin/env python
import unittest
import os
import numpy as np
import astropy.io.fits as pyfits
from test.test_helper import TestHelper
from plico.utils.snapshotable import Snapshotable


class SnapshotableTest(unittest.TestCase):
    TEMP_FITS = "./tmp.fits"

    def setUp(self):
        if os.path.exists(self.TEMP_FITS):
            os.remove(self.TEMP_FITS)

    def test_prefix_prepending(self):
        snapshot = {"AB": 1, "CD": 2}
        Snapshotable.prepend("FOO", snapshot)
        self.assertEqual(2, len(list(snapshot.keys())))
        self.assertEqual(1, snapshot["FOO.AB"])
        self.assertEqual(2, snapshot["FOO.CD"])

    def test_prepending_supports_chaining(self):
        snapshot = {"A": 42}
        result = Snapshotable.prepend("FOO", snapshot)
        self.assertEqual(snapshot, result)

    def test_FITSHeader_conversion(self):
        snapshot = {"A": 3, "B": 3.5, "C": False, "D": "urga"}
        hdr = Snapshotable.asFITSHeader(snapshot)
        restoredSnapshot = Snapshotable.fromFITSHeader(hdr)
        self.assertEqual(snapshot, restoredSnapshot)

    def test_skips_FITS_header_value_for_NoneType(self):
        hdr = Snapshotable.asFITSHeader({"A": None})
        self.assertEqual({}, Snapshotable.fromFITSHeader(hdr))

    def _checkSnapshotKeyRestore(self, KEY, VALUE, snapshot):
        hdr = Snapshotable.asFITSHeader(snapshot)
        pyfits.writeto(self.TEMP_FITS, np.array([1]), hdr)
        restoredHdr = pyfits.getheader(self.TEMP_FITS)
        restoredSnapshot = Snapshotable.fromFITSHeader(restoredHdr)
        self.assertEqual(VALUE, restoredSnapshot[KEY])

    def test_longest_uncutted_FITS_card_image(self):
        key = "K" * 59
        value = "V" * 8
        self._checkSnapshotKeyRestore(key, value, {key: value})

    def test_FITS_card_with_max_key_length_and_overlong_value_is_cutted(self):
        key = "K" * 59
        value = "1234567890"
        self._checkSnapshotKeyRestore(key, "12345678", {key: value})

    def test_FITS_card_image_with_overlong_value_is_cutted(self):
        key = "K" * 9
        value = "V" * 200
        self._checkSnapshotKeyRestore(key, value[0:58], {key: value})

    @unittest.skip("Do we understand FITS standard?")
    def test_FITS_card_image_with_max_key_length_and_overlong_stringified_float_value(self):
        key = "K" * 59
        value = 1 / 3.
        self._checkSnapshotKeyRestore(key, value, {key: value})

    def test_remove_items_with_value_None(self):
        d = {'a': 3, 'b': None, 'c': "Ceh"}
        Snapshotable.removeEntriesWithValueNone(d)
        TestHelper.areDictEqual({'a': 3, 'c': "Ceh"}, d)


if __name__ == "__main__":
    unittest.main()




class DummySockets(object):

    def __init__(self, prefix='dummySocket'):
        self._prefix= prefix

    def serverSubscriber(self):
        return "%s-%s" % (self._prefix, 'subscriber')

    def serverSubscriberAddress(self):
        return "%s-%s" % (self._prefix, 'subscriberAddress')

    def serverRequest(self):
        return "%s-%s" % (self._prefix, 'request')

    def serverRequestAddress(self):
        return "%s-%s" % (self._prefix, 'requestAddress')

    def serverStatus(self):
        return "%s-%s" % (self._prefix, 'status')

    def serverStatusAddress(self):
        return "%s-%s" % (self._prefix, 'statusAddress')

    def serverDisplay(self):
        return "%s-%s" % (self._prefix, 'display')

    def serverDisplayAddress(self):
        return "%s-%s" % (self._prefix, 'displayAddress')




import appdirs
from pkg_resources import Requirement, resource_filename
import shutil
import os
from plico.utils.addtree import mkdirp

try:
    from pathlib import Path
except ImportError:
    from pathlib2 import Path  # python 2 backport


__version__= "$Id: config_file_manager.py 50 2018-04-26 21:42:20Z lbusoni $"


class ConfigFileManager():

    def __init__(self, appName, appAuthor, pythonPackageName):
        self._appName= appName
        self._appAuthor= appAuthor
        self._packageName= pythonPackageName
        self._appdirs= appdirs.AppDirs(self._appName, self._appAuthor)


    def getConfigFilePath(self):
        confPath= os.path.join(self._appdirs.user_config_dir,
                               '%s.conf' % self._packageName)
        return confPath


    def _getConfigFilePathInPackage(self):
        return resource_filename(
            Requirement(self._packageName),
            "%s/conf/%s.conf" % (self._packageName, self._packageName))


    def doesConfigFileExists(self):
        return os.path.isfile(self.getConfigFilePath())


    def installConfigFileFromPackage(self, overwrite=False):
        if self.doesConfigFileExists() and (overwrite is False):
            return
        source= self._getConfigFilePathInPackage()
        dest= self.getConfigFilePath()
        mkdirp(os.path.dirname(self.getConfigFilePath()))
        shutil.copyfile(source, dest)



import threading
import time
from plico.utils.decorator import override, synchronized, returns
from plico.utils.convergeable import Convergeable


class FakeConvergeable(Convergeable):

    def __init__(self):
        Convergeable.__init__(self)
        self._gain = 0.0
        self._convergenceStepInvocationCnt = 0
        self._measureConvergenceInvocationCnt = 0
        self._stepSleepDurationSec = 0
        self._mtx = threading.Lock()
        self._condition = threading.Condition()
        self._stepBlockingEnabled = False
        self._exceptionRequested = False
        self._converged = False

    @override
    def name(self):
        return "FakeConvergeable"

    def setAsConverged(self):
        self._converged = True

    def setAsUnconverged(self):
        self._converged = False

    def _panicIfRequested(self):
        if self._exceptionRequested:
            raise Exception("Exception is desired")

    @override
    def performOneConvergenceStep(self):
        self._incrementStepCount()
        self._panicIfRequested()
        time.sleep(self._stepSleepDurationSec)
        self._blockIfAny()

    @synchronized("_mtx")
    @override
    def measureConvergence(self):
        self._measureConvergenceInvocationCnt += 1
        self._panicIfRequested()

    def getMeasureConvergenceCount(self):
        return self._measureConvergenceInvocationCnt

    def requestPanic(self):
        self._exceptionRequested = True

    @synchronized("_condition")
    def unblockStep(self):
        self._condition.notifyAll()
        self._stepBlockingEnabled = False

    @synchronized("_condition")
    def blockStep(self):
        self._stepBlockingEnabled = True

    @synchronized("_condition")
    def _blockIfAny(self):
        if self._stepBlockingEnabled:
            self._condition.wait()

    def setStepSleepDurationSec(self, durationSec):
        self._stepSleepDurationSec = durationSec

    @synchronized("_mtx")
    def _incrementStepCount(self):
        self._convergenceStepInvocationCnt += 1

    @synchronized("_mtx")
    def getConvergenceStepCount(self):
        return self._convergenceStepInvocationCnt

    @returns(bool)
    def hasConverged(self):
        return self._converged



import numpy as np
from math import factorial




class ZernikeGenerator(object):
    """ZernikeGenerator

    A class to generate Zernike polynomials and gradient of Zernike
    polynomials

    Example:
        >>> numberOfPixelOnPupilDiameter= 64
        >>> zg= ZernikeGenerator(numberOfPixelOnPupilDiameter)
        >>> asti= zg.getZernike(5)
        >>> asti[
        >>> type(asti)
        numpy.ma.core.MaskedArray
        >>> asti.shape
        (64,64)
        >>> dTiltdx= zg.getDerivativeX(2)
        >>> dTiltdx.mean()
        2.0
        >>> dTiltdx.ptp()
        0.0

    """

    def __init__(self, nPixelOnDiameter):
        self._nPixel= nPixelOnDiameter
        self._rhoMap, self._thetaMap= self._polar_array(self._nPixel)
        self._dx= None
        self._dy= None
        self._dictCache= {}
        self._dictDxCache= {}
        self._dictDyCache= {}


    def getRadius(self):
        return self._nPixel / 2.


    def _derivativeCoeffX(self, index):
        if (self._dx is None) or (self._dx.shape[0] < index):
            self._dx= self._computeDerivativeCoeffX(index)
        return self._dx[0:index, 0:index]


    def _derivativeCoeffY(self, index):
        if (self._dy is None) or (self._dy.shape[0] < index):
            self._dy= self._computeDerivativeCoeffY(index)
        return self._dy[0:index, 0:index]


    def _degree(self, index):
        n= int(0.5 * (np.sqrt(8 * index - 7) - 3)) + 1
        cn= n * (n + 1) / 2 + 1
        if n % 2 == 0:
            m = int(index - cn + 1) // 2 * 2
        else:
            m = int(index - cn) // 2 * 2 + 1
        radialDegree= n
        azimuthalFrequency= m
        return radialDegree, azimuthalFrequency


    def _rnm(self, radialDegree, azimuthalFrequency, rhoArray):
        n= radialDegree
        m= azimuthalFrequency
        rho= rhoArray
        if (n - m) %2 != 0:
            raise Exception("n-m must be even. Got %d-%d" % (n, m))
        if abs(m)>n:
            raise Exception("The following must be true |m|<=n. Got %d, %d" %
                            (n, m))
        mask=np.where(rho <= 1, False, True)

        if(n==0 and m==0):
            return np.ma.masked_array(data=np.ones(rho.shape), mask=mask)
        rho= np.where(rho < 0, 0, rho)
        Rnm= np.zeros(rho.shape)
        S= (n - abs(m)) // 2
        for s in range(0, S + 1):
            CR= pow(-1, s) * factorial(n - s) / \
                (factorial(s) * factorial(-s + (n + abs(m)) / 2) *
                 factorial(-s + (n - abs(m)) / 2))
            p= CR * pow(rho, n - 2 * s)
            Rnm= Rnm + p
        return np.ma.masked_array(data=Rnm, mask=mask)



    def _polar(self, index, rhoArray, thetaArray):
        n, m= self._degree(index)
        rho= rhoArray
        theta= thetaArray

        Rnm= self._rnm(n, m, rho)
        NC= np.sqrt(2 * (n + 1))
        if m == 0:
            return np.sqrt(0.5)* NC * Rnm
        if index % 2 == 0:
            return NC* Rnm * np.cos(m * theta)
        else:
            return NC * Rnm * np.sin(m * theta)


    def _polar_array(self, nPixel):
        X, Y= np.mgrid[-1+ 1. / nPixel: 1- 1. / nPixel: nPixel * 1j,
                       -1+ 1. / nPixel: 1- 1. / nPixel: nPixel * 1j]
        r = np.sqrt(X**2 + Y**2)
        th= np.arccos(np.transpose(X * 1./ r))
        th= np.where(th < 2.* np.pi, th, 0)
        th= np.where(X< 0, 2.* np.pi- th, th)
        return r, th


    def getZernikeDict(self, indexVector):
        ret= {}
        for index in indexVector:
            ret[index]= self.getZernike(index)
        return ret


    def getZernike(self, index):
        """getZernike

        Retrieve a map representing the index-th Zernike polynomial

        Args:
            index (int): The index of Zernike map to be generated,
                following Noll 1976 ordering.

        Returns:
            np.array: A map representing the index-th Zernike polynomial

        """
        if index not in list(self._dictCache.keys()):
            self._dictCache[index]= self._polar(index, self._rhoMap,
                                                self._thetaMap)
        return self._dictCache[index]


    def __getitem__(self, index):
        return self.getZernike(index)


    def _computeDerivativeCoeffX(self, index):
        jmax= index
        G_mat=np.zeros((jmax, jmax))
        for i in range(1, jmax + 1):
            for j in range(1, jmax + 1):
                ni, mi= self._degree(i)
                nj, mj= self._degree(j)
                if (
                    (
                        (
                            (
                                mi != 0 and mj != 0
                            ) and (
                                (
                                    _isEven(i) and _isEven(j)
                                ) or (
                                    _isOdd(i) and _isOdd(j)
                                )
                            )
                        ) or (
                            (
                                (mi == 0) and _isEven(j)
                            ) or (
                                (mj == 0) and _isEven(i)
                            )
                        )
                    ) and (
                        (mj == mi + 1) or (mj == mi - 1)
                    ) and (
                        j < i
                    )
                ):
                    G_mat[i - 1, j - 1]= np.sqrt((ni + 1) * (nj + 1))
                    if ((mi == 0) or (mj == 0)):
                        G_mat[i - 1, j - 1] *= np.sqrt(2)
        return G_mat


    def _computeDerivativeCoeffY(self, index):
        jmax= index
        G_mat=np.zeros((jmax, jmax))
        for i in range(1, jmax + 1):
            for j in range(1, jmax + 1):
                ni, mi= self._degree(i)
                nj, mj= self._degree(j)
                if (
                    (
                        (
                            (
                                mi != 0 and mj != 0
                            ) and (
                                (
                                    _isOdd(i) and _isEven(j)
                                ) or (
                                    _isEven(i) and _isOdd(j)
                                )
                            )
                        ) or (
                            (
                                (mi == 0) and _isOdd(j)
                            ) or (
                                (mj == 0) and _isOdd(i)
                            )
                        )
                    ) and (
                        (mj == mi + 1) or (mj == mi - 1)
                    ) and (
                        j < i
                    )
                ):
                    G_mat[i - 1, j - 1]= np.sqrt((ni + 1) * (nj + 1))
                    if ((mi == 0) or (mj == 0)):
                        G_mat[i - 1, j - 1] *= np.sqrt(2)
                    if (
                        (
                            (
                                (mj == mi + 1) and _isOdd(i)
                            ) or (
                                (mj == mi - 1) and _isEven(i)
                            )
                        ) and (
                            mi != 0
                        )
                    ):
                        G_mat[i - 1, j - 1] *= -1

        return G_mat


    def getDerivativeXDict(self, indexVector):
        ret= {}
        for index in indexVector:
            ret[index]= self.getDerivativeX(index)
        return ret


    def getDerivativeYDict(self, indexVector):
        ret= {}
        for index in indexVector:
            ret[index]= self.getDerivativeY(index)
        return ret


    def getDerivativeX(self, index):
        if index not in list(self._dictDxCache.keys()):
            self._dictDxCache[index]= self._computeDerivativeX(index)
        return self._dictDxCache[index]


    def _computeDerivativeX(self, index):
        coeffX= self._derivativeCoeffX(index)
        dx= self.getZernike(1) * 0.
        for i in range(1, index):
            dx += coeffX[index - 1, i - 1] * self.getZernike(i)
        return dx


    def getDerivativeY(self, index):
        if index not in list(self._dictDyCache.keys()):
            self._dictDyCache[index]= self._computeDerivativeY(index)
        return self._dictDyCache[index]


    def _computeDerivativeY(self, index):
        coeffY= self._derivativeCoeffY(index)
        dy= self.getZernike(1) * 0.
        for i in range(1, index):
            dy += coeffY[index - 1, i - 1] * self.getZernike(i)
        return dy


def _isOdd(num):
    return num % 2 != 0


def _isEven(num):
    return num % 2 == 0


